// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "scene/SceneItemComponent.hpp"
#include "scene/components/display/IRenderable.hpp"
#include "display/RenderPipeline.hpp"
#include "display/RenderManager.hpp"
#include "UICanvas.hpp"
#include "util/mathutils.hpp"

namespace Dawn {
  enum UIComponentAlign {
    UI_COMPONENT_ALIGN_START,
    UI_COMPONENT_ALIGN_MIDDLE,
    UI_COMPONENT_ALIGN_END,
    UI_COMPONENT_ALIGN_STRETCH
  };

  enum UIComponentAlignUnit {
    UI_COMPONENT_ALIGN_UNIT_SCALE,
    UI_COMPONENT_ALIGN_UNIT_PERCENT
  };

  class UIComponent : public SceneItemComponent, public UIComponentDimensional {
    protected:
      float_t width = 1;
      float_t height = 1;
      
      /**
       * Simply returns this UI Components' dimensional parent, used for the
       * alignment of this item.
       * 
       * @return Pointer to the parent dimensional.
       */
      UIComponentDimensional * getParentDimensional();

      /**
       * Internal method to update the alignment of this item.
       */
      virtual void updateAlignment();
      
    public:
      StateProperty<bool_t> alignmentNeedsUpdating;
      StateEvent<> eventAlignmentUpdated;

      /**
       * Method used to calculate alignment values.
       * 
       * @param alignmentValue Alignment value.
       * @param parentSize Parent size.
       * @param unit Alignment unit.
       * @return The calculated alignment value.
       */
      static float_t calculateAlignmentValue(
        float_t alignmentValue,
        float_t parentSize,
        enum UIComponentAlignUnit unit
      );

      /**
       * Method used to calculate alignment dimensions.
       * 
       * @param align Alignment  value enumator.
       * @param unitStart Alignment start unit.
       * @param unitEnd Alignment end unit.
       * @param position Output position floating point.
       * @param size Output size floating point.
       * @param outerSize Outer size (of the parent).
       * @param innerSize Inner size (of this element's content).
       * @param alignment Alignment settings.
       */
      static void calculateDimensions(
        enum UIComponentAlign align,
        enum UIComponentAlignUnit unitStart,
        enum UIComponentAlignUnit unitEnd,
        float_t *position,
        float_t *size,
        float_t outerSize,
        float_t innerSize,
        glm::vec2 alignment
      );

      // @optional
      StateProperty<enum UIComponentAlign> alignX;
      // @optional
      StateProperty<enum UIComponentAlign> alignY;
      // @optional
      StateProperty<enum UIComponentAlignUnit> alignUnitLeft;
      // @optional
      StateProperty<enum UIComponentAlignUnit> alignUnitTop;
      // @optional
      StateProperty<enum UIComponentAlignUnit> alignUnitRight;
      // @optional
      StateProperty<enum UIComponentAlignUnit> alignUnitBottom;
      // @optional
      StateProperty<glm::vec4> alignment;

      UIComponent(SceneItem *item);

      /**
       * Returns the canvas that this UI element is belonging to.
       * 
       * @return Pointer to the UI Canvas this component is a child of.
       */
      virtual UICanvas * getCanvas();

      float_t getWidth() override;
      float_t getHeight() override;
      float_t getChildOffsetX() override;
      float_t getChildOffsetY() override;
      void onStart() override;

      friend class UICanvas;
  };
}